package model;

import java.awt.Color;
import java.awt.Point;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.nio.ByteBuffer;
import java.util.Objects;

public class Player implements Serializable {
    public final static int representationSize = 6; // length of the packed byte array 
    public final Color color;
    public final Point pos, smoothPos, scaledPos;
    public Direction moveDir, faceDir, nextDir;
    public int speed;
    
    private final Point startPos;
    private final int mouthSpeed = 10;
    private final int maxOpenMouth = 40;
    private final int scale;
    private int mouth;
    private boolean mouthOpens;

    public Player(int x, int y, int scale, Color color, int speed) {
        pos = new Point(x, y);
        startPos = new Point(pos);
        scaledPos = new Point(x * scale, y * scale);
        smoothPos = new Point(scaledPos);
        this.mouth = maxOpenMouth;
        this.mouthOpens = true;
        this.moveDir = Direction.STAY;
        this.faceDir = Direction.RIGHT;
        this.nextDir = Direction.STAY;
        this.color = color;
        this.scale = scale;
        this.speed = speed;
    }

    public Player(Player p) {
        pos = new Point(p.pos);
        startPos = new Point(p.startPos);
        scaledPos = new Point(p.scaledPos);
        smoothPos = new Point(p.smoothPos);
        mouth = p.mouth;
        mouthOpens = p.mouthOpens;
        moveDir = p.moveDir;
        faceDir = p.faceDir;
        nextDir = p.nextDir;
        color = p.color;
        scale = p.scale;
        speed = p.speed;        
    }
    
    public void updateFromStream(InputStream is) throws IOException{
        // Representation requires 2 + 2 + 1 + 1 = 6 bytes
        DataInputStream dis = new DataInputStream(is);
        int sx = dis.readShort();
        int sy = dis.readShort();
        pos.setLocation(sx / scale, sy / scale);
        scaledPos.setLocation(pos.x * scale, pos.y * scale);        
        smoothPos.setLocation(sx, sy);
        this.mouth = dis.readByte();
        this.faceDir = Direction.values()[dis.readByte()];
    }
    
    public byte[] toByteArray(){
        // Representation requires 2 + 2 + 1 = 6 bytes
        // May write instead into a DataOutputStream(new ByteArrayOutputStream())
        ByteBuffer bb = ByteBuffer.allocate(representationSize);
        bb.putShort((short)smoothPos.x);
        bb.putShort((short)smoothPos.y);
        bb.put((byte)mouth);
        bb.put((byte)faceDir.ordinal());
        return bb.array();
    }
    
    public void animateMouth(){
        if (mouthOpens){
            mouth += mouthSpeed;
        } else {
            mouth -= mouthSpeed;
        }
        if (mouth >= maxOpenMouth) mouthOpens = false;
        if (mouth <= mouthSpeed) mouthOpens = true;
    }
    
    public int getMouth() {
        return mouth;
    }
    
    public void setSmoothX(int sx){
        smoothPos.x = sx;
        pos.x = sx / scale;
        scaledPos.x = pos.x * scale;
    }
    
    public void setSmoothY(int sy){
        smoothPos.y = sy;
        pos.y = sy / scale;
        scaledPos.y = pos.y * scale;
    }
    
    public void setSmoothCoord(int sx, int sy){
        setSmoothX(sx);
        setSmoothY(sy);
    }

    public void setNextDir(Direction d){
        nextDir = d;
    }
    
    public void reset(){
        pos.setLocation(startPos);
        scaledPos.setLocation(pos.x * scale, pos.y * scale);
        smoothPos.setLocation(scaledPos);
        moveDir = Direction.STAY;
        faceDir = Direction.RIGHT;
        nextDir = Direction.STAY;
        mouth = maxOpenMouth;
        mouthOpens = true;
    }

    @Override
    public String toString() {
        return "Player{" + "color=" + color + 
                ", pos=" + pos + 
                ", smoothPos=" + smoothPos + 
                ", scaledPos=" + scaledPos + 
                ", moveDir=" + moveDir + 
                ", faceDir=" + faceDir + 
                ", nextDir=" + nextDir + 
                ", speed=" + speed + 
                ", startPos=" + startPos + 
                ", mouthSpeed=" + mouthSpeed + 
                ", maxOpenMouth=" + maxOpenMouth + 
                ", scale=" + scale + 
                ", mouth=" + mouth + 
                ", mouthOpens=" + mouthOpens + '}';
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Player other = (Player) obj;
        if (this.speed != other.speed) return false;
        if (this.mouthSpeed != other.mouthSpeed) return false;
        if (this.maxOpenMouth != other.maxOpenMouth) return false;
        if (this.scale != other.scale) return false;
        if (this.mouth != other.mouth) return false;
        if (this.mouthOpens != other.mouthOpens) return false;
        if (this.moveDir != other.moveDir) return false;
        if (this.faceDir != other.faceDir) return false;
        if (this.nextDir != other.nextDir) return false;
        if (!Objects.equals(this.color, other.color)) return false;
        if (!Objects.equals(this.pos, other.pos)) return false;
        if (!Objects.equals(this.smoothPos, other.smoothPos)) return false;
        if (!Objects.equals(this.scaledPos, other.scaledPos)) return false;
        if (!Objects.equals(this.startPos, other.startPos)) return false;
        return true;
    }
}